/*
 * JSVG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jsvg;

import com.jsevy.jxml.*;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.io.StringWriter;
import java.util.StringTokenizer;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;


/**
 * Class representing a Scalable Vector Graphics document, which owns an svgGraphics on which drawing commands can be made. The document's
 * toSVGString method can then be called to generate the SVG text corresponding to the drawing done on the Graphics object.
 * 
 * The typical workflow is as follows:
 * 
 * <pre>
 * 
 * // Create an svg document and get its associated SVGGraphics instance
 * SVGDocument svgDocument = new SVGDocument(); 
 * SVGGraphics svgGraphics = svgDocument.getGraphics(); 
 * 
 * // Do drawing commands as on any other Graphics. If you have a paint(Graphics) method, 
 * // you can just use it with the SVGGraphics instance since it's a subclass of Graphics. 
 * graphics.setColor(Color.RED);  
 * graphics.setStroke(new BasicStroke(3));
 * graphics.drawLine(0, 0, 1000, 500); 
 * graphics.drawRect(1000, 500, 150, 150); 
 * graphics.drawRoundRect(20, 200, 130, 100, 20, 10); 
 * 
 * // Get the flat SVG output as a string - it's just XML text - and  save  in a file for use 
 * // with LibreOffice Draw or any other draw package that accepts flat SVG (.svg) files
 * String stringOutput = svgDocument.toSVGString(); 
 * String filePath = "path/to/file.svg"; 
 * FileWriter fileWriter = new FileWriter(filePath); 
 * fileWriter.write(svgText); 
 * fileWriter.flush(); 
 * fileWriter.close();
 * 
 * </pre>
 * 
 * @author jsevy
 *
 */
public class SVGDocument extends XMLDocument
{
    
    //private Vector<SVGEntity> entities;
    
    private SVGGraphics graphics;
    
    private SVGStyles styles;
    
    
    private Rectangle2D.Double documentBounds = new Rectangle2D.Double(0, 0, 10, 10);

    private boolean autosize = true;
    
    public static String svgNamespaceURI = "http://www.w3.org/2000/svg";
    
    
    /**
     * Create a new SVGDocument
     */
    public SVGDocument()
    {
        super("svg", svgNamespaceURI);
        
        this.graphics = new SVGGraphics(this);
        
        //this.entities = new Vector<SVGEntity>();
        this.styles = new SVGStyles();
        
        populateXML();
    }
    
    
    
    
    /** Set the units for interpreting the dimension values in the SVG file.
     *
     *  @param unitsCode  SVG code indicating units:
     *      0 = Unitless
     *      1 = Inches
     *      2 = Feet
     *      3 = Miles
     *      4 = Millimeter
     *      5 = Centimeters
     *      6 = Meters
     *      7 = Kilometers
     *      8 = Microinches
     *      9 = Mils
     *      10 = Yards
     * 
     */
    public void setUnits(int unitsCode)
    {
        String[] unitsMatrix = 
            {
            /* 0 = */ "",
            /* 1 = */ "in",
            /* 2 = */ "ft",
            /* 3 = */ "mi",
            /* 4 = */ "mm",
            /* 5 = */ "cm",
            /* 6 = */ "m",
            /* 7 = */ "km",
            /* 8 = */ "uin",
            /* 9 = */ "mils",
            /* 10 = */ "yd",
        };
        
        SVGEntity.units = unitsMatrix[unitsCode];
    }
    
    
    
    /**
     * Set the document width and height, which equates to the overall viewBox width and height for the SVG 
     * drawing added as an attribute to the top-level svg element. Note that the units used are those set with a 
     * call to setUnits(). This also turns off autosizing of the document.
     * 
     * @param documentWidth    Document/viewBox width, in units set by call to setUnits()
     * @param documentHeight   Document/viewBox height, in units set by call to setUnits()
     */
    public void setDocumentSize(double documentWidth, double documentHeight)
    {
        this.documentBounds.width = documentWidth;
        this.documentBounds.height = documentHeight;
        
        this.autosize = false;
    }
    
    
    /**
     * Set the document rectangle, which equates to the viewBox for the SVG drawing which
     * is added as an attribute to the top-level svg element. Note that the units used are those set 
     * with a call to setUnits(). This also turns off autosizing of the document.
     * 
     * @param bounds    Document/viewBox rectangle, in units set by call to setUnits()
     */
    public void setDocumentBounds(Rectangle2D.Double bounds)
    {
        this.documentBounds = bounds;
        
        this.autosize = false;
    }
    
    
    /**
     * Return a copy of the rectangle defining the viewport of the document.
     * 
     * @return  A copy of the rectangle defining the document viewport
     */
    public Rectangle2D.Double getDocumentBounds()
    {
        return new Rectangle2D.Double(documentBounds.x, documentBounds.y, documentBounds.width, documentBounds.height);
    }
    
    
    /**
     * Specify whether or not to autosize the document, which will set the viewport to the minimum size needed to
     * enclose all of the graphics entities. The default is true, but this may be set to false by a call to
     * this method, at which point a default viewport will be used, or by a call to either the setDocumentSize() or 
     * setDocumentBounds() methods, which explicitly specify the document (viewport) size.
     * 
     * @param autosize    Boolean specifying whether the document should be automatically sized to enclose all
     *                    of the graphical entities in the generated SVG; default is true
     */
    public void autosizeDocument(boolean autosize)
    {
        this.autosize  = autosize;
    }
    
    
    
    /**
     * Return the smallest rectangle which encloses the graphical elements included in the drawing up to the point of this call.
     * This can be used to specify a viewport for the SVG, for example if it is desired to include margins around the 
     * graphics, by calling after drawing is completed and using the result to define an enlarged rectangle for use in 
     * a call to setDocumentBounds().
     * 
     * @return  Rectangle enclosing all of the graphical elements in the design at the point of the call to this method
     */
    public Rectangle2D.Double getEntityBounds()
    {
        Rectangle2D.Double bounds = new Rectangle2D.Double();
        
        // run through entities to find max and min
        XMLElement documentElement = (XMLElement)this.getDocumentElement();
        XMLNodeList svgEntityList = (XMLNodeList)documentElement.getChildNodes();
        if (svgEntityList.size() > 0)
        {
            SVGEntity entity = (SVGEntity)svgEntityList.elementAt(0);
            Rectangle2D.Double entityBounds = (Rectangle2D.Double)entity.getBounds();
            double xMin = entityBounds.x;
            double yMin = entityBounds.y;
            double xMax = entityBounds.x + entityBounds.width;
            double yMax = entityBounds.y + entityBounds.height;
            
            for (int i = 1; i < svgEntityList.size(); i++)
            {
                entity = (SVGEntity)svgEntityList.elementAt(i);
                entityBounds = (Rectangle2D.Double)entity.getBounds();

                xMin = Math.min(xMin, entityBounds.x);
                yMin = Math.min(yMin, entityBounds.y);
                xMax = Math.max(xMax, entityBounds.x + entityBounds.width);
                yMax = Math.max(yMax, entityBounds.y + entityBounds.height);
                
            }
            
            bounds.x = xMin;
            bounds.y = yMin;
            bounds.width = xMax - xMin;
            bounds.height = yMax - yMin;
        }
        
        return bounds;
    }
    
    
    
    /**
     * Set the scale factor for dimensions in the Java drawing - how many Java units/pixels
     * correspond to one unit in the SVG output. Note that the units can be set with a call to 
     * setUnits
     * 
     * @param javaUnitsPerUnit  How many Java units/pixels correspond to one SVG unit
     */
    public void setScaleFactor(double javaUnitsPerUnit)
    {
        SVGEntity.scaleFactor = javaUnitsPerUnit;
    }
    
    
    
    
    /**
     * Get the SVGGraphics associated with this document for use with standard Graphics drawing operations to generate
     * an svg text representation.
     * 
     * @return	The SVGGraphics associated with this document, on which standard Java Graphics drawing calls can be made
     */
    public SVGGraphics getGraphics()
    {
        return this.graphics;
    }
    
    
    /**
     * Get a graphics style corresponding to the current Graphics2D instance, and whether the shape
     * should be stroked and/or filled. Will return an existing corresponding style if there is one,
     * or create a new one and add it to the list of global styles if needed.
     * 
     * @param graphics  Java Graphics2D instance
     * @param stroked   True if shape is to be stroked with current color of supplied graphics instance
     * @param filled    True if shape is to be filled with current color of supplied graphics instance
     * @return          Corresponding SVGGraphicsStyle
     */
    protected SVGGraphicsStyle getGraphicsStyle(Graphics2D graphics, boolean stroked, boolean filled)
    {
        return styles.getGraphicsStyle(graphics, stroked, filled);
    }
    
    
    /**
     * Get a text style corresponding to the current Graphics2D instance. Will return an existing corresponding
     * style if there is one, or create a new one and add it to the list of global styles if needed.
     * 
     * @param graphics  Java Graphics2D instance
     * @return          Corresponding SVGTextStyle
     */
    protected SVGTextStyle getTextStyle(Graphics2D graphics)
    {
        return styles.getTextStyle(graphics);
    }
    
    
    private void populateXML()
    {
        // namespace attribute added in document creation
        //this.getDocumentElement().setAttribute("xmlns:", svgNamespaceURI);
    }
    
    
    
    /**
     * Return the SVG text associated with this SVG document. This includes the header and
     * entities populated with content generated by graphics calls on the associated SVGGraphics.
     * 
     * @return The SVG text associated with this document.
     * @throws TransformerException 
     */
    public String toSVGString()
    {
        // first, add viewbox to set user scale; wait until XML generation so the units and page size can be set if desired
        if (autosize) 
        {
            // set the document size to enclose all graphics elements
            Rectangle2D.Double bounds = this.getEntityBounds();
            this.setDocumentBounds(bounds);
        }
        
        this.getDocumentElement().setAttribute("width", SVGObject.addUnits(documentBounds.width));
        this.getDocumentElement().setAttribute("height", SVGObject.addUnits(documentBounds.height));
        this.getDocumentElement().setAttribute("viewBox", documentBounds.x + " " + documentBounds.y +" " + documentBounds.width + " " + documentBounds.height);
        
        DOMSource domSource = new DOMSource(this);
        StringWriter stringWriter = new StringWriter();
        
        try
        {
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer transformer = tf.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            
            transformer.transform(domSource, new StreamResult(stringWriter));
        }
        catch (TransformerException e)
        {
            // really shouldn't ever happen...
            System.out.println("TransformerException while creating SVG XML: " + e.getMessage());
        }
        
        //xmlBuffer.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + "\n");
        //xmlBuffer.append(this.toXMLString(0, 2));
        
        StringBuffer xmlBuffer = stringWriter.getBuffer();
        return xmlBuffer.toString();
    }
    
    
    /**
     * Utility method used by the associated SVGGraphics object.
     * 
     * @param entity an svgEntity instance
     */
    public void addEntity(SVGEntity entity)
    {
        //entities.add(entity);
        this.getDocumentElement().appendChild(entity);
    }
    
    
    /**
     * Utility method for use with addSet method of SVGObject to enable animation of elements.
     * 
     * @return  Last entity added to the document; may be null if no entities
     */
    public SVGEntity getLastEntity()
    {
        //entities.add(entity);
        return (SVGEntity)this.getDocumentElement().getLastChild();
    }
    
    
    
    /**
     * For the given SVG entity, set the value of the specified attribute to the value specified by toValue, beginning 
     * at the specified start time and for the specified duration.
     * 
     * @param entity            the entity whose attribute is to be set
     * @param attributeName     name of attribute to be set
     * @param toValue           value to which attribute should be set; this may be a numerical value (primitive or
     *                          object), a String, or a Java Color object. A numerical value will be scaled by the current
     *                          scale value as set by setScale(); a Color object will be converted into the RGB format used
     *                          in SVG; and a String will be included as-is
     * @param beginTimeSeconds  when the attribute value should be set, in seconds
     * @param durationSeconds   how long the attribute should remain set, in seconds, or the String "indefinite"
     * @param scaleValue       whether to scale the from and to values by the current scale value as set by setScale(); only
     *                          meaningful if the values are numerical
     */
    protected void setAttribute(SVGEntity entity, String attributeName, Object toValue, double beginTimeSeconds, Object durationSeconds, boolean scaleValue)
    {
        if (entity != null)
        {
            String begin = Double.toString(beginTimeSeconds);
            
            // handles the case where "indefinite" is supplied
            String duration = durationSeconds.toString();
            
            // determine what type of value is being supplied, and use that 
            // to determine how to turn it into a string
            String toValueString = svgStringifyAttributeValue(toValue, scaleValue);
            
            // add SVGAnimate to entity
            entity.addSVGSet(attributeName, toValueString, begin, duration);
        }
    }
    
    
    /**
     * For the last added SVG entity (the last object created by one the Graphics draw methods), set the value of the 
     * specified attribute to the value specified by toValue, beginning at the specified start time
     * and for the specified duration.
     * 
     * Internally, this adds an SVG <set> object as a child of the supplied entity. The exact syntax and behavior can
     * be found in the SVG specification.
     * 
     * @param attributeName     name of attribute to be set
     * @param toValue           value to which attribute should be set; this may be a numerical value (primitive or
     *                          object), a String, or a Java Color object. A numerical value will be scaled by the current
     *                          scale value as set by setScale(); a Color object will be converted into the RGB format used
     *                          in SVG; and a String will be included as-is
     * @param beginTimeSeconds  when the attribute value should be set
     * @param durationSeconds   how long the attribute should remain set, in seconds, or the String "indefinite"
     * @param scaleValue       whether to scale the from and to values by the current scale value as set by setScale(); only
     *                          meaningful if the values are numerical
     */
    public void setAttribute(String attributeName, Object toValue, double beginTimeSeconds, Object durationSeconds, boolean scaleValue)
    {
        SVGEntity entity = getLastEntity();
        this.setAttribute(entity, attributeName, toValue, beginTimeSeconds, durationSeconds, scaleValue);
    }
    
    
    
    /**
     * For the supplied SVG entity, animate the value of the specified attribute between the supplied fromValue and
     * toValue, starting at the specified beginning time and over the period specified by the duration. Repeat the indicated
     * number of times; if freezeAtEnd is true, hold the final value of the attribute, otherwise remove the attribute
     * changes when the animation period is finished.
     * 
     * Internally, this adds an SVG <animate> object as a child of the supplied entity. The exact syntax and behavior can
     * be found in the SVG specification.
     * 
     * @param entity            the entity to which the animation should be added
     * @param attributeName     name of the attribute to be animated
     * @param fromValue         value to which attribute should be set to start the animation; this may be a numerical value
     *                          (primitive or object), a String, or a Java Color object. A numerical value will be scaled 
     *                          by the current scale value if scaleValue is true; a Color object will be converted into the 
     *                          RGB format used in SVG; and a String will be included as-is. The special value "" indicates that
     *                          the starting value should be whatever the current value is for the attribute.
     * @param toValue           value to which attribute should approach over the animation duration; this may be a numerical value
     *                          (primitive or object), a String, or a Java Color object. A numerical value will be scaled 
     *                          by the current scale value if scaleValue is true; a Color object will be converted into the 
     *                          RGB format used in SVG; and a String will be included as-is.
     * @param beginTimeSeconds  when the animation should begin, in seconds
     * @param durationSeconds   the duration of the animation (time from starting to ending attribute value), in seconds
     * @param repeatCount       an integer indicating how many times the animation should repeat, or the String "indefinite"
     * @param freezeAtEnd       if true, the animation freezes in its last state; if false, the effects of the animation
     *                          are removed after all cycles have been completed
     * @param scaleValues       whether to scale the from and to values by the current scale value as set by setScale(); only
     *                          meaningful if the values are numerical
     */
    protected void animateAttribute(SVGEntity entity, String attributeName, Object fromValue, Object toValue, double beginTimeSeconds, Object durationSeconds, Object repeatCount, boolean freezeAtEnd, boolean scaleValues)
    {
        if (entity != null)
        {
            String begin = Double.toString(beginTimeSeconds);
            
            // handles the case where "indefinite" is supplied
            String duration = durationSeconds.toString();
            
            // handles the case where "indefinite" is supplied
            String repeat = repeatCount.toString();
            
            String fill;
            if (freezeAtEnd)
                fill = "freeze";
            else
                fill = "remove";
            
            // determine what type of value is being supplied, and use that 
            // to determine how to turn it into a string
            String fromValueString = svgStringifyAttributeValue(fromValue, scaleValues);
            String toValueString = svgStringifyAttributeValue(toValue, scaleValues);
            
            // add SVGAnimate to entity
            entity.addSVGAnimate(attributeName, fromValueString, toValueString, begin, duration, repeat, fill);
            
            
            /*
            // add SVGAnimate to entity:
            // we have a special case, in which toValue is a semicolon-separated list of values; if there's
            // a semicolon, use the variant constructor
            if (toValueString.contains(";"))
            {
                entity.addSVGAnimate(attributeName, toValueString, begin, duration, repeat, fill);
            }
            else
            {
                entity.addSVGAnimate(attributeName, fromValueString, toValueString, begin, duration, repeat, fill);
            }
            */
            
            
        }
    }
    
    
    /**
     * For the supplied SVG entity, animate the value of the specified attribute between the supplied fromValue and
     * toValue, starting at the specified beginning time and over the period specified by the duration. Repeat the indicated
     * number of times; if freezeAtEnd is true, hold the final value of the attribute, otherwise remove the attribute
     * changes when the animation period is finished.
     * 
     * Internally, this adds an SVG <animate> object as a child of the supplied entity. The exact syntax and behavior can
     * be found in the SVG specification.
     * 
     * @param attributeName     name of the attribute to be animated
     * @param fromValue         value to which attribute should be set to start the animation; this may be a numerical value
     *                          (primitive or object), a String, or a Java Color object. A numerical value will be scaled 
     *                          by the current scale value if scaleValue is true; a Color object will be converted into the 
     *                          RGB format used in SVG; and a String will be included as-is. The special value "" indicates that
     *                          the starting value should be whatever the current value is for the attribute.
     * @param toValue           value to which attribute should approach over the animation duration; this may be a numerical value
     *                          (primitive or object), a String, or a Java Color object. A numerical value will be scaled 
     *                          by the current scale value if scaleValue is true; a Color object will be converted into the 
     *                          RGB format used in SVG; and a String will be included as-is.
     * @param beginTimeSeconds  when the animation should begin, in seconds
     * @param durationSeconds   the duration of the animation (time from starting to ending attribute value), in seconds
     * @param repeatCount       an integer indicating how many times the animation should repeat, or the String "indefinite"
     * @param freezeAtEnd       if true, the animation freezes in its last state; if false, the effects of the animation
     *                          are removed after all cycles have been completed
     * @param scaleValues       whether to scale the from and to values by the current scale value as set by setScale(); only
     *                          meaningful if the values are numerical
     */
    public void animateAttribute(String attributeName, Object fromValue, Object toValue, double beginTimeSeconds, Object durationSeconds, Object repeatCount, boolean freezeAtEnd, boolean scaleValues)
    {
        SVGEntity entity = getLastEntity();
        this.animateAttribute(entity, attributeName, fromValue, toValue, beginTimeSeconds, durationSeconds, repeatCount, freezeAtEnd, scaleValues);
    }
    
    
    /**
     * For the supplied SVG entity, animate the value of the specified attribute between the supplied fromValue and
     * toValue, starting at the specified beginning time and over the period specified by the duration. Repeat the indicated
     * number of times; if freezeAtEnd is true, hold the final value of the attribute, otherwise remove the attribute
     * changes when the animation period is finished.
     * 
     * Internally, this adds an SVG <animate> object as a child of the supplied entity. The exact syntax and behavior can
     * be found in the SVG specification.
     * 
     * @param entity            the entity to which the animation should be added
     * @param attributeName     name of the attribute to be animated
     * @param valueList         string with semicolon-separated entries; these will be interpreted as a sequence of values the attribute
     *                          should be animated through over the duration of the animation, as per the SVG specification.
     *                          Numerical values will be scaled by the current scale value if scaleValue is true
     * @param beginTimeSeconds  when the animation should begin, in seconds
     * @param durationSeconds   the duration of the animation (time from starting to ending attribute value), in seconds
     * @param repeatCount       an integer indicating how many times the animation should repeat, or the String "indefinite"
     * @param freezeAtEnd       if true, the animation freezes in its last state; if false, the effects of the animation
     *                          are removed after all cycles have been completed
     * @param scaleValues       whether to scale the from and to values by the current scale value as set by setScale(); only
     *                          meaningful if the values are numerical
     */
    protected void animateAttribute(SVGEntity entity, String attributeName, String valueList, double beginTimeSeconds, Object durationSeconds, Object repeatCount, boolean freezeAtEnd, boolean scaleValues)
    {
        if (entity != null)
        {
            String begin = Double.toString(beginTimeSeconds);
            
            // handles the case where "indefinite" is supplied
            String duration = durationSeconds.toString();
            
            // handles the case where "indefinite" is supplied
            String repeat = repeatCount.toString();
            
            String fill;
            if (freezeAtEnd)
                fill = "freeze";
            else
                fill = "remove";
            
            // determine what type of value is being supplied, and use that 
            // to determine how to turn it into a string
            String valueListString = processValueList(valueList, scaleValues);
            
            // add SVGAnimate to entity:
            // we have a special case, in which toValue is a semicolon-separated list of values; if there's
            // a semicolon, use the variant constructor
            entity.addSVGAnimate(attributeName, valueListString, begin, duration, repeat, fill);
            
        }
    }
    
    
    
    /**
     * For the supplied SVG entity, animate the value of the specified attribute between the supplied fromValue and
     * toValue, starting at the specified beginning time and over the period specified by the duration. Repeat the indicated
     * number of times; if freezeAtEnd is true, hold the final value of the attribute, otherwise remove the attribute
     * changes when the animation period is finished.
     * 
     * Internally, this adds an SVG <animate> object as a child of the supplied entity. The exact syntax and behavior can
     * be found in the SVG specification.
     * 
     * @param attributeName     name of the attribute to be animated
     * @param valueList         string with semicolon-separated entries; these will be interpreted as a sequence of values the attribute
     *                          should be animated through over the duration of the animation, as per the SVG specification.
     *                          Numerical values will be scaled by the current scale value if scaleValue is true
     * @param beginTimeSeconds  when the animation should begin, in seconds
     * @param durationSeconds   the duration of the animation (time from starting to ending attribute value), in seconds
     * @param repeatCount       an integer indicating how many times the animation should repeat, or the String "indefinite"
     * @param freezeAtEnd       if true, the animation freezes in its last state; if false, the effects of the animation
     *                          are removed after all cycles have been completed
     * @param scaleValues       whether to scale the from and to values by the current scale value as set by setScale(); only
     *                          meaningful if the values are numerical
     */
    public void animateAttribute(String attributeName, String valueList, double beginTimeSeconds, Object durationSeconds, Object repeatCount, boolean freezeAtEnd, boolean scaleValues)
    {
        SVGEntity entity = getLastEntity();
        this.animateAttribute(entity, attributeName, valueList, beginTimeSeconds, durationSeconds, repeatCount, freezeAtEnd, scaleValues);
    }
    
    
    
    private String processValueList(String valueList, boolean scaleValue)
    {
        String processedList = new String();
        
        // for now, just scale any numerical entries
        StringTokenizer st = new StringTokenizer(valueList, ";");
        
        if (st.countTokens() > 0)
        {
            while (st.hasMoreTokens()) 
            {
                String valueString = st.nextToken();
                
                try 
                {
                    Double value = new Double(valueString);
                    
                    // it's numerical; scale it if needed
                    processedList += svgStringifyAttributeValue(value, scaleValue) + ";";
                }
                catch(NumberFormatException e)
                {
                    // not numerical; include as is
                    processedList += valueString + ";";
                }
            }
            
            // remove last semicolon
            processedList = processedList.substring(0, processedList.lastIndexOf(';'));
        }
        
        return processedList;
    }
    
    
    
    /**
     * Get a sensible value for the supplied object for use in <set> and <animate> objects
     * 
     * @param value     the value being used in a set or animate call
     * @return          a (hopefully) SVG-friendly String representation
     */
    private String svgStringifyAttributeValue(Object value, boolean scaleValue)
    {
        String valueString;
        
        if (value instanceof String)
        {
            valueString = (String)value;
        }
        else if (value instanceof Number)
        {
            if (scaleValue)
            {
                double doubleValue = ((Number)value).doubleValue();
                valueString = SVGObject.scaleValue(doubleValue);
            }
            else
            {
                valueString = ((Number)value).toString();
            }
            
        }
        else if (value instanceof Color)
        {
            Color color = (Color)value;
            valueString = "#" + (Integer.toHexString(color.getRGB()).substring(2));
        }
        else
        {
            // almost certainly not what should be here...
            valueString = value.toString();
        }
        
        return valueString;
    }
    
    
    /**
     * Utility routine for use with addSetToEntity method when geometric values are supplied. 
     * Given a double value, scale it according to the scale factor previously set, and add the units string.
     * @param value     Value to be scaled
     * @return          String giving the value with units appended.
     */
    public static String scaleValueWithUnits(double value)
    {
        return SVGObject.scaleValueWithUnits(value);
    }
    
    
    /**
     * Utility routine for use with addSetToEntity method when geometric values are supplied. Given a double value, 
     * scale it according to the scale factor previously set, and return the String representation.
     * @param value     Value to be scaled
     * @return          Scaled value
     */
    public static String scaleValue(double value)
    {
        return SVGObject.scaleValue(value);
    }
    
    
    /**
     * Utility routine for use with addSetToEntity method when geometric values are supplied. 
     * Given a double value, convert to a String and add the units string.
     * @param value     Value
     * @return          String giving the value with units appended.
     */
    public static String addUnits(double value)
    {
        return SVGObject.addUnits(value);
    }
    
}